/*****************************************************
Chip type               : ATmega8L
AVR Core Clock frequency: 1,000000 MHz
*****************************************************/

#include <mega8.h>
#include <STDLIB.h>

#define A PORTB.4
#define B PORTC.4
#define C PORTB.0
#define D PORTB.2
#define E PORTB.3
#define F PORTB.5
#define G PORTD.5
#define DP PORTB.1
#define A0 PORTD.4
#define A1 PORTD.3
#define A2 PORTD.2
#define A3 PORTD.1
#define A4 PORTD.0
#define A5 PORTC.5

#define button0 PINC.1    //mode
#define button1 PINC.2    //dn
#define button2 PINC.3    //up

#define buzzer PORTC.0

bit t_trigger, beep_trigger, sleep_trigger;
char display_buffer [6], display_counter;
eeprom signed char c_corr;  
signed char c_corr_tmp_updn, c_corr_tmp;

enum btn_states 
{b_prs, b_rls, b_hold} btn0, btn1, btn2;
enum time_modes 
{s_adj, m_adj, h_adj, normal_operation,  corr}mode = normal_operation;

void ports_init ()
{
PORTB=0b00000000;
DDRB=  0b00111111;

PORTC=0b01001110;
DDRC=  0b00110001;

PORTD=0b00000000;
DDRD=  0b11111111;
}

void t2_init ()
{
// Timer/Counter 2 initialization
// Mode: Normal top=0xFF

ASSR |= 1<<AS2; // Clock source: TOSC1 pin
TCCR2 = 0b00000011;   // Clock value: PCK2/32  (0,25s) 

TIMSK = 0x40;
}

struct time
{
 char quarter_seconds;
 signed char dpt;
 signed char hours;
 signed char minutes;
 signed char seconds;
 } systime, time_adj, time_tmp;

void upd_time (struct time *t);
void convert_time (struct time *t, struct time *k);
void clear_display ();
void display_time ();
void adj_time (struct time *t, struct time *p);
void skan_keys ();

interrupt [TIM2_OVF] void t2_ovf_isr(void)
{
 upd_time (&systime);
}

void beep ()
{
static unsigned int bsw;

if (beep_trigger) 
{
if (!bsw) {bsw = 400; buzzer = ~buzzer;}                          
bsw--; 
}
else {buzzer = 0;}
}

void main(void)
{
ports_init ();
t2_init ();

c_corr_tmp_updn = c_corr;

ACSR &= ~(1<<ACD);
MCUCR |= 1<<SE;
#asm("sei")

while (1)
      {
        skan_keys (); 
        adj_time (&systime, &time_adj);
        beep ();
        while (ACSR.ACO)     //power loss
            {
            DDRB = 0; DDRC = 0; DDRD = 0;
            PORTB = 0;PORTC = 0; PORTD = 0;
            sleep_trigger = 1;
            #asm ("sleep")
            }
        if (sleep_trigger) {ports_init (); sleep_trigger = 0;} 
      }
}

void upd_time (struct time *t)
{       
static bit corr_trigger;

t->quarter_seconds ++; t_trigger = 0;
if (t->quarter_seconds > 3) {t->seconds ++; t->quarter_seconds = 0; t->dpt = ~t->dpt; beep_trigger = 0;}
if (t->seconds > 59) {t->seconds = 0; t->minutes ++;} 
if (t->minutes >59) {t->minutes = 0; t->hours ++; beep_trigger = 1;}
if (t->hours >23) {t->hours = 0; corr_trigger = 1;}

if (corr_trigger) 
{
 if ((!t->seconds)&&(c_corr>0)) {t->seconds = t->seconds + c_corr; corr_trigger = 0;}
 if ((t->seconds == 59)&&(c_corr<0)) {t->seconds = t->seconds + c_corr; corr_trigger = 0;} 
}
}

void convert_time (struct time *t, struct time *k)            
{
  char i;
  k->seconds = t->seconds;
  k->minutes = t->minutes;
  k->hours = t->hours;
  
  for (i=0; i<2; i++) {display_buffer [i] = k->seconds % 10; k->seconds /= 10;}
  for (i=2; i<4; i++) {display_buffer [i] = k->minutes % 10; k->minutes /= 10;}
  for (i=4; i<6; i++) {display_buffer [i] = k->hours % 10; k->hours /= 10;}
}  

void clear_display () 
{ 
A=1; B=1; C=1; D=1; E=1; F=1; G=1; DP=1; A0 = 0; A1 = 0; A2 = 0; A3 = 0; A4 = 0; A5 = 0;
}

void display_time ()
{

 if (display_counter<5) {display_counter++;} else {display_counter = 0;}
        
 clear_display ();

 switch (display_buffer[display_counter])
 {
  case 0: A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 1: A=1; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 2: A=0; B=0; C=1; D=0; E=0; F=1; G=0; break;
  case 3: A=0; B=0; C=0; D=0; E=1; F=1; G=0; break;
  case 4: A=1; B=0; C=0; D=1; E=1; F=0; G=0; break;
  case 5: A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 6: A=0; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 7: A=0; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 8: A=0; B=0; C=0; D=0; E=0; F=0; G=0; break;
  case 9: A=0; B=0; C=0; D=0; E=1; F=0; G=0; break; 
  case 10: A=1; B=1; C=1; D=0; E=0; F=0; G=1; break;     //L
  case 11: A=1; B=0; C=0; D=1; E=0; F=0; G=0; break;     //H
  case 12: A=1; B=1; C=1; D=1; E=1; F=1; G=1; break;     //
  default : A=1; B=1; C=1; D=1; E=1; F=1; G=0; break;    //-
   }
 
 switch (display_counter)
 {
  case 0: A0 = 1; break;
  case 1: A1 = 1; break;
  case 2: A2 = 1; DP = 0; break;     //if (systime.dpt) {DP = 0;}
  case 3: A3 = 1; break;
  case 4: A4 = 1; DP = 0; break;
  case 5: A5 = 1; break; 
  default: break;
 }
  
}

void skan_keys ()
{
static unsigned int count0, count1, count2, count3;

if (button0){ btn0 = b_rls;}
if (button1){ btn1 = b_rls;}
if (button2){ btn2 = b_rls;}

if (!button0) {count0++; } 
else {count0 = 0;}
if (!button1) {count1++; } 
else {count1 = 0;}
if (!button2) {count2++; } 
else {count2 = 0;}

if ((count0 > 155)&&(btn0 != b_hold)) {btn0 = b_prs; count0 = 0; count3++;}
if ((count1 > 155)&&(btn1 != b_hold)) {btn1 = b_prs; count1 = 0; count3++;} 
if ((count2 > 155)&&(btn2 != b_hold)) {btn2 = b_prs; count2 = 0; count3++;}

if ((count3 > 25)&&(btn0 == b_prs)) {btn0 = b_hold; count3 = 0;}
if ((count3 > 25)&&(btn1 == b_prs)) {btn1 = b_hold; count3 = 0;}
if ((count3 > 25)&&(btn2 == b_prs)) {btn2 = b_hold; count3 = 0;}
}

void adj_time (struct time *t, struct time *p)                  
{
char n;
static bit trigger, trigger1, trigger2, trigger3, 
ee_trigger, tadj_trigger_s, tadj_trigger_m, tadj_trigger_h;

if (button2) {trigger3 = 0;}
if (button1) {trigger2 = 0;}
if (btn0 == b_rls) {trigger = 0; trigger1 = 0;}
if ((btn0 == b_prs)&&(mode != corr)) {if (!trigger) {trigger = 1; if (mode<normal_operation) {mode++; } else {mode = s_adj;}}}
if ((btn0 == b_hold)&&(!trigger1)) {trigger1 = 1; if (mode != corr) {mode = corr; } else {mode = normal_operation;}}


 switch (mode)
 {
  case s_adj: 
  if (btn2 == b_hold) {if (time_adj.seconds < 59) { if(!t_trigger) time_adj.seconds++; t_trigger = 1;} else {time_adj.seconds = 0;}}
  if (btn1 == b_hold) {if (time_adj.seconds) { if(!t_trigger) time_adj.seconds--; t_trigger = 1;} else {time_adj.seconds = 59;}}
   
  if ((btn2 == b_prs)&&(!trigger3)) {trigger3 = 1; if (time_adj.seconds < 59) {time_adj.seconds++;} else {time_adj.seconds = 0;}}
  if ((btn1 == b_prs)&&(!trigger2)) {trigger2 = 1; if (time_adj.seconds) {time_adj.seconds--;} else {time_adj.seconds = 59;}}
  convert_time (&time_adj, &time_tmp); for (n=2; n<6; n++) {display_buffer[n] = 12;}
  display_time ();
  tadj_trigger_s = 0;
  break;
  
  case m_adj:   
  if (btn2 == b_hold) {if (time_adj.minutes < 59) { if(!t_trigger) time_adj.minutes++; t_trigger = 1;} else {time_adj.minutes = 0;}}
  if (btn1 == b_hold) {if (time_adj.minutes) { if(!t_trigger) time_adj.minutes--; t_trigger = 1;} else {time_adj.minutes = 59;}}
  
  if ((btn2 == b_prs)&&(!trigger3)) {trigger3 = 1; if (time_adj.minutes< 59) {time_adj.minutes++;} else {time_adj.minutes = 0;}}
  if ((btn1 == b_prs)&&(!trigger2)) {trigger2 = 1; if (time_adj.minutes) {time_adj.minutes--;} else {time_adj.minutes = 59;}}
  convert_time (&time_adj, &time_tmp); 
  for (n=0; n<2; n++) {display_buffer[n] = 12;} for (n=4; n<6; n++) {display_buffer[n] = 12;}  
  display_time ();
  tadj_trigger_m =0;
  break;
  
  case h_adj:
  if (btn2 == b_hold) {if (time_adj.hours < 23) { if(!t_trigger) time_adj.hours++; t_trigger = 1;} else {time_adj.hours = 0;}}
  if (btn1 == b_hold) {if (time_adj.hours) { if(!t_trigger) time_adj.hours--; t_trigger = 1;} else {time_adj.hours = 23;}}    
     
  if ((btn2 == b_prs)&&(!trigger3)) {trigger3 = 1; if (time_adj.hours < 23) {time_adj.hours++;} else {time_adj.hours = 0;}}
  if ((btn1 == b_prs)&&(!trigger2)) {trigger2 = 1; if (time_adj.hours) {time_adj.hours--;} else {time_adj.hours = 23;}}  
  convert_time (&time_adj, &time_tmp); 
  for (n=0; n<4; n++) {display_buffer[n] = 12;}                  
  display_time ();
  tadj_trigger_h = 0;
  break;
  //------------------------------------------------
  case corr:   
  if ((btn2 == b_prs)&&(!trigger3)) {trigger3 = 1; if (c_corr_tmp_updn < 59) {c_corr_tmp_updn++;} else {c_corr_tmp_updn = 0;}}
  if ((btn1 == b_prs)&&(!trigger2))  {trigger2 = 1; if (c_corr_tmp_updn > -59) {c_corr_tmp_updn--;} else {c_corr_tmp_updn = 0;}}
  
  c_corr_tmp = cabs (c_corr_tmp_updn);
  for (n=0; n<2; n++) {display_buffer [n] = c_corr_tmp % 10; c_corr_tmp /= 10;}
  for (n=2; n<5; n++) {display_buffer[n] = 12;} 
  if (c_corr_tmp_updn>0) {display_buffer[5] = 11;} 
  else if (c_corr_tmp_updn<0) {display_buffer[5] = 10;}
  else {display_buffer[5] = 13;}
  display_time ();
  ee_trigger = 0;
  break;  

// --------------------------------------------------- &systime, &time_adj     tp
  case normal_operation: 
  if (!ee_trigger) {ee_trigger = 1; c_corr = c_corr_tmp_updn;}
  if (!tadj_trigger_s) {tadj_trigger_s = 1; t->seconds = p->seconds;}
  if (!tadj_trigger_m) {tadj_trigger_m = 1; t->minutes = p->minutes;}  
  if (!tadj_trigger_h) {tadj_trigger_h = 1; t->hours = p->hours;}
    
  convert_time (&systime, &time_tmp); display_time ();
  break; 
  
  default: break; 
 }


}